enum Comparable_ { LessThan, Equal, GreaterThan }

// interface Comparable<T> { // Similar to Java "functional interface"
//     (t: T): Comparable_; // Signature
// }
// // Compiler error: "Type 'Giraffe' provides no match for the signature '<T>(t: T): Comparable_'" (no solution exists!)
// class Giraffe implements Comparable<Giraffe> {
//     constructor(readonly _name: string, readonly _height: number) {
//     }
// }

abstract class Animal {
}

abstract class Mammal extends Animal {
}

abstract class Oviparous extends Animal {
}

// class Ornithorhynchus extends Mammal, Oviparous { // Not supported as in C++!
// }

class Giraffe extends Mammal {
    constructor(readonly _name: string, readonly _height: number) {
        super(); // Mandatory
    }
}

interface Comparator<T> {
    compare: (t1: T, t2: T) => Comparable_; // 'compare' has type '(t1: T, t2: T) => Comparable_'
}

class Zoo implements Comparator<Giraffe> {
    compare = (g1: Giraffe, g2: Giraffe): Comparable_ => {
        if (g1._height < g2._height) return Comparable_.LessThan;
        if (g1._height === g2._height) return Comparable_.Equal;
        console.assert(g1._height > g2._height);
        return Comparable_.GreaterThan;
    };
    // compare = function (g1: Giraffe, g2: Giraffe): Comparable_ {
    //     if (g1._height < g2._height) return Comparable_.LessThan;
    //     if (g1._height === g2._height) return Comparable_.Equal;
    //     console.assert(g1._height > g2._height);
    //     return Comparable_.GreaterThan;
    // };
}

interface Comparable<T> { // Java functional interface
    compareTo(t: T): Comparable_;
}

class Elephant extends Mammal implements Comparable<Elephant> {
    constructor(readonly _name: string, readonly _weight: number) {
        super();
    }

    compareTo(e: Elephant): Comparable_ { // 'public' by default
        if (this._weight < e._weight) return Comparable_.LessThan;
        if (this._weight === e._weight) return Comparable_.Equal;
        console.assert(this._weight > e._weight);
        return Comparable_.GreaterThan;
    }
}

class Sort_illustration<T extends Comparable<T>> {
    private readonly _implementation: Array<T> = new Array/*<T>*/(); // Don't forget to fill in somewhere else!

    sort(left: number, right: number): void { // Naive (slow) sort
        console.assert(left < right);
        for (let i: number = left; i < right; i++) {
            for (let j: number = i + 1; j <= right; j++) {
                if (this._implementation[i].compareTo(this._implementation[j]) === Comparable_.GreaterThan) {
                    let temp: T = this._implementation[i];
                    this._implementation[i] = this._implementation[j];
                    this._implementation[j] = temp;
                }
            }
        }
    }
}

let t_toString = <T>(t: T) => (<Object>t).toString();

let identity = <T>(t: T): T => { // <=> 'function identity<T>(t: T): T {'
    return t;
}

let identity_alias: <U>(u: U) => U = identity; // Type of 'identity' and 'identity_alias' is '<T>(t: T) => T' <=> '<U>(u: U) => U'

interface Main { // Similar to Java "functional interface" -> "function type" in TypeScript
    (): void; // Call signature
}

class Human_being {
    protected readonly _birth_date: number;

    constructor(readonly birth_date: string) {
        this._birth_date = Date.parse(birth_date);
    }

    // private constructor(readonly birth_date: number) { // No other constructor is accepted, even 'private'!
    //     this._birth_date = birth_date;
    // }

    // cloning(): this { // This differs from 'Current' polymorphic type in Eiffel!
    //     return this; // Only 'this' object complies with 'this' *TYPE*!
    // }

    cloning(): Human_being {
        console.log("cloning in Human_being...\n");
        // return new Human_being(this._birth_date); // No way because multiple constructors aren't accepted!
        return new Human_being(new Date(this._birth_date).toDateString());
    }

    that_s_me(): this { // This differs from 'Current' polymorphic type in Eiffel!
        return this; // Only 'this' object complies with 'this' *TYPE*!
    }
}

class Female extends Human_being {

    cloning(): Female {
        console.log("cloning in Female...\n");
        return new Female(new Date(this._birth_date).toDateString());
    }
}

const main: Main = () => {
    console.log("Zero: " + t_toString(0));
    let result: string = identity("Franck Barbier"); // <=> 'identity<string>("Franck Barbier")'
    result = identity_alias("Franck Barbier");
    const Babar: Elephant = new Elephant("Babar", 100), Jumbo: Elephant = new Elephant("Jumbo", 200);

    const Eve: Human_being = new Female("0, 0, 0");
    console.assert(Eve.that_s_me().constructor === Female); // 'true' is displayed!
    const Eve_clone: Human_being = Eve.cloning(); // 'cloning in Female...' is displayed => covariance applies in TypeScript!
};
main();



